// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test8mat.cpp
// Purpose:	Time Measurements for Rule Implementations
// Last Change:	03.02.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Example Program 8:
// p(X, Y) :- q(X, Y).
// p(X, Z) :- q(X, Y), p(Y, Z).
//
// q := {(i,i+1)|i = 1,...,1000}


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES = 1000;

const bool PRINT_EXTENSION = false;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


// ----------------------------------------------------------------------------
// Predicate p - implemented as linked list
// ----------------------------------------------------------------------------

class p {
	friend class p_cursor;
	friend class p_cursor_diff;
public:
	int col_1() const {return Col_1; }
	int col_2() const {return Col_2; }

	static void insert(int c1, int c2)
	{
		p *new_elem = new p(c1, c2);
		if(Last)
			Last->Next = new_elem;
		else
			First = new_elem;
		Last = new_elem;
	}

	static bool next_iter()
	{
		if(Iter_Last) {
			Iter_First = Iter_Last->Next;
			Iter_Last = Last;
		}
		else {
			Iter_First = First;
			Iter_Last = Last;
		}
		return(Iter_First != 0);
	}

private:
	p(int c1, int c2)
	{
		Col_1 = c1;
		Col_2 = c2;
		Next = 0;
	}

	int Col_1;
	int Col_2;
	p *Next;

	static class p* First;
	static class p* Last;
	static class p* Iter_First;
	static class p* Iter_Last;
};

class p* p::First = 0;
class p* p::Last = 0;
class p* p::Iter_First = 0;
class p* p::Iter_Last = 0;


class p_cursor {
public:
	p_cursor()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
	}

	bool fetch()
	{
		if(Before_First) {
			Current = p::First;
			Before_First = false;
		}
		else
			Current = Current->Next;
		return(Current != NULL);
	}

	int col_1() { return Current->col_1(); }
	int col_2() { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p*   Current;
};


class p_cursor_diff {
public:
	p_cursor_diff()
	{
		Before_First = true;
		Current = 0;
	}

	void open()
	{
		Before_First = true;
		Current = p::Iter_First;
		Last = p::Iter_Last;
	}

	bool fetch()
	{
		if(Before_First) {
			Before_First = false;
			return(Current != NULL);
		}
		else if(Current != Last) {
			Current = Current->Next;
			return true;
		}
		else
			return false;
	}

	int col_1() { return Current->col_1(); }
	int col_2() { return Current->col_2(); }

	void close() {}

private:
	bool Before_First;
	p*   Current;
	p*   Last;
};


//-----------------------------------------------------------------------------
// Predicate q: Tuples generated by computation
//-----------------------------------------------------------------------------

class q_cursor {
public:
	q_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < Q_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i+1; }

	void close() {}

private:
	int i;
};


//-----------------------------------------------------------------------------
// Rule Implementation: Materialization
//-----------------------------------------------------------------------------

void exec_rule1()
{
	// p(X, Y) :- q(X, Y).
	q_cursor q_c;
	q_c.open();
	while(q_c.fetch()) {
		int X = q_c.col_1();
		int Y = q_c.col_2();
		p::insert(X, Y);
	}
}

void exec_rule2()
{
	// p(X, Z) :- q(X, Y), p(Y, Z).
	while(p::next_iter()) {
		q_cursor q_c;
		q_c.open();
		while(q_c.fetch()) {
			int X = q_c.col_1();
			int Y = q_c.col_2();
			p_cursor_diff p_c;
			p_c.open();
			while(p_c.fetch()) {
				if(p_c.col_1() == Y) {
					int Z = p_c.col_2();
					p::insert(X, Z);
				}
			}
		}
	}
}

void exec_prog()
{
	// Execute initialization rule:
	exec_rule1();
	//
	// Execute recursive rule:
	exec_rule2();
}

//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1 = 0;
	int c2 = 0;
	int num_tuples = 0;

	cout << "Test of Method 1: Materialization [Program 8]\n";
	t.start();
	exec_prog();
	p_cursor p_c;
	p_c.open();
	while(p_c.fetch()) {
		c1 += p_c.col_1();
		c2 += p_c.col_2();
		num_tuples++;
	}
	p_c.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p:\n\n";
		p_c.open();
		while(p_c.fetch()) {
			cout << "p(" << p_c.col_1() << ", " << p_c.col_2()
				<< ").\n";
		}
		p_c.close();
	}

	cout << "\n\nTotal number of tuples: " << num_tuples << "\n";
	cout << "Column sums: col1=" << c1 << ", col2=" << c2 << "\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}


